#include "../SDK/foobar2000.h"
#include "../foosion_helpers/helpers.h"

#include <history.h>
#include "config.h"
#include "history_loader.h"

#ifdef _DEBUG
#	define CONSOLE_OUTPUT(s,x) console::output(s,x)
#else
#	define CONSOLE_OUTPUT(s,x) /* nothing */
#endif

#define CONSOLE_INFO(x) CONSOLE_OUTPUT(console::SEVERITY_INFO,x)
#define CONSOLE_WARNING(x) CONSOLE_OUTPUT(console::SEVERITY_WARNING,x)
#define CONSOLE_ERROR(x) CONSOLE_OUTPUT(console::SEVERITY_CRITICAL,x)

//////////////////////////////////////////////////
// #define's for experimental features
//////////////////////////////////////////////////


//////////////////////////////////////////////////
// #define's for debugging features
//////////////////////////////////////////////////

//#define VERBOSE_INDEX_ADJUSTMENT

static bool initialized = false; // prevents history_callbacks being called while loading history


//////////////////////////////////////////////////
// helper functions for titleformatting
//////////////////////////////////////////////////

/**
 * Computes number of days between two dates.
 * The computation ignores the time of day, so the
 * difference between 2004-08-05 23:45:21 (b) and
 * 2004-08-06 00:05:17 (a) is 1.
 */
static int get_days_ago(FILETIME * a, FILETIME * b) {
	// 100 nanoseconds intervals per day
	static const __int64 INTERVALS_PER_DAY = ((__int64)(24*60*60)*(__int64)10000000);
	__int64 d_a = (*(__int64 *)a) / INTERVALS_PER_DAY;
	__int64 d_b = (*(__int64 *)b) / INTERVALS_PER_DAY;
	__int64 diff = d_a - d_b;
	return (int)diff;
}

static int get_days_ago(SYSTEMTIME * a, SYSTEMTIME * b) {
	FILETIME ft_a;
	FILETIME ft_b;
	SystemTimeToFileTime(a, &ft_a);
	SystemTimeToFileTime(b, &ft_b);
	return get_days_ago(&ft_a, &ft_b);
}

static int get_hours_ago(FILETIME * a, FILETIME * b) {
	// 100 nanoseconds intervals per hour
	static const __int64 INTERVALS_PER_HOUR = ((__int64)(60*60)*(__int64)10000000);
	/*
	__int64 h_a = (*(__int64 *)a) / INTERVALS_PER_HOUR;
	__int64 h_b = (*(__int64 *)b) / INTERVALS_PER_HOUR;
	*/
	__int64 diff = ((*(__int64 *)a) - (*(__int64 *)b)) / INTERVALS_PER_HOUR;
	return (int)diff;
}


//////////////////////////////////////////////////
// helper class for playlist tracking
//////////////////////////////////////////////////

class playlist_handle : public service_base {
private:
	unsigned index;

protected:
	playlist_handle(int index) {
		this->index = index;
	}

public:
	unsigned get_index() {
		return index;
	}

	void set_index(unsigned p_index) {
		index = p_index;
	}

	bool is_deleted() {
		return get_index() == (unsigned)-1;
	}

	static playlist_handle * create(unsigned index = -1) {
		return new service_impl_p1_t< playlist_handle, int >(index);
	}
};


//////////////////////////////////////////////////
// history_handle implementation
//////////////////////////////////////////////////

class history_handle_i : public history_handle {
private:
	long ref_count;
	metadb_handle * track;
	packed_date timestamp;
	playlist_handle * the_playlist;

	virtual ~history_handle_i() {
		if (track) track->handle_release();
		if (the_playlist) the_playlist->service_release();
		issued--;
	}

public:
	static unsigned issued;

	history_handle_i(metadb_handle * p_track, const packed_date & p_timestamp, playlist_handle * p_playlist = 0) {
		issued++;
		ref_count = 1;
		track = p_track;
		timestamp = p_timestamp;
		the_playlist = p_playlist;
		if (track) track->handle_add_ref();
		if (the_playlist) the_playlist->service_add_ref();
	}

	virtual void handle_add_ref() {
		interlocked_increment(&ref_count);
	}

	virtual void handle_release() {
		int rv = interlocked_decrement(&ref_count);
		if (rv <= 0) delete this;
	}

	virtual metadb_handle * handle_get_track() {return track;}

	virtual void handle_get_timestamp(packed_date & out) {
		out = timestamp;
	}

	virtual int handle_compare_timestamp(history_handle * p_src) {
		history_handle_i * src = (history_handle_i *)p_src;
		return timestamp.compare(src->timestamp);
	}

	virtual void handle_format_title(string_base & out, const char * spec, const char * p_extra_items) {
		extra_items_helper extra_items(p_extra_items);

#ifdef _DEBUG
		if (the_playlist && the_playlist->get_index() != (unsigned)-1) {
			extra_items.add_item("history_playlist", the_playlist->get_index());
		}
#endif

		TIME_ZONE_INFORMATION tzi;
		int dst = GetTimeZoneInformation(&tzi);

		if (dst != TIME_ZONE_ID_STANDARD) {
			// bias is in minutes
			int bias = -(tzi.Bias + ((dst == TIME_ZONE_ID_STANDARD) ? tzi.StandardBias : tzi.DaylightBias));
			char temp[8];
			if (bias < 0) {
				sprintf(temp, "-%02d:%02d", (-bias)/60, (-bias)%60);
			} else {
				sprintf(temp, "+%02d:%02d", bias/60, bias%60);
			}
			extra_items.add_item("history_timezone", temp);
			extra_items.add_item("history_timezone_minutes", bias);
		}

		packed_date d;
		handle_get_timestamp(d);
		SYSTEMTIME st_utc;
		d.get(st_utc);
		extra_items.add_item("history_timestamp", (const char *)string_print_date_utc(st_utc));

		FILETIME ft_utc;
		SystemTimeToFileTime(&st_utc, &ft_utc);
		FILETIME ft_local;
		FileTimeToLocalFileTime(&ft_utc, &ft_local);
		SYSTEMTIME st_local;
		FileTimeToSystemTime(&ft_local, &st_local);
		string_print_date timestamp(st_local);
		extra_items.add_item("history_day_of_week", st_local.wDayOfWeek+1);
		extra_items.add_item("history_date", &timestamp[0], 10);
		extra_items.add_item("history_time", &timestamp[11], 8);
		FILETIME ft_now;
		GetSystemTimeAsFileTime(&ft_now);
		FILETIME ft_now_local;
		FileTimeToLocalFileTime(&ft_now, &ft_now_local);
		int days_ago = get_days_ago(&ft_now_local, &ft_local);
		extra_items.add_item("history_days_ago", days_ago);
		int hours_ago = get_hours_ago(&ft_now_local, &ft_local);
		extra_items.add_item("history_hours_ago", hours_ago);
		
		handle_get_track()->handle_format_title(out, spec, extra_items);
	}

	playlist_handle * get_playlist() {
		return the_playlist;
	}
};

unsigned history_handle_i::issued = 0;


//////////////////////////////////////////////////
// history service implementation
//////////////////////////////////////////////////

class history_i : public history {
private:
	critical_section2 cs;
	history_handle_list data;
	int now_playing, cursor;
	// every index > bound_next is "in the future", bound_next+1 is the "next in history" track
	// every index < bound_prev is "in the past", bound_prev-1 is the "Previous in history" track
	int bound_next, bound_prev;

	void fire_item_added(int index);
	void fire_items_added(int start, int count);
	void fire_item_modified(int index);
	void fire_item_replaced(int index);
	void fire_playing_track_changed(int from, int to);

	ptr_list_t< playlist_handle > playlists;
	unsigned pl_index;

	ptr_list_t< history_callback > callbacks;

	history_handle_i * get_item_locked(int index);

	history_handle_i * create_handle(metadb_handle * track, const packed_date * timestamp, playlist_handle * p_playlist);

public:
	history_i() {
		now_playing = -1;
		bound_next = -1;
		bound_prev = 0;

		cursor = 0;
		
		pl_index = (unsigned)-1;
	}

	~history_i() {
		data.delete_all();
		if (history_handle_i::issued) {
			uMessageBox(0, string_printf("history_handle leaks, %d objects", history_handle_i::issued), "Error", 
				MB_OK);
		}
	}

	void add_track(metadb_handle * track, const packed_date * timestamp = 0, playlist_handle * p_playlist = 0);
	void add_track_autoplaylist(metadb_handle * track, const packed_date * timestamp = 0);

	virtual int get_count();

	virtual history_handle * get_item(int index);
	virtual void get_data(ptr_list_interface<history_handle> & out);
	virtual void remove_mask(const bit_array & items);

	virtual void format_title(int index, string_base & out, const char * spec, const char * extra_items);

	virtual int get_now_playing();
	virtual void play_item(int index);

	virtual int history_lock() {return cs.enter();}
	virtual int history_try_lock() {return cs.try_enter();}
	virtual int history_unlock() {return cs.leave();}

	virtual void set_playback_cursor(int index);
	virtual int get_playback_cursor();

	virtual int get_index_from_offset(int offset);
	virtual int get_offset_from_index(int index);

	virtual void add_callback(history_callback * ptr, bool callonadd);
	virtual void remove_callback(history_callback * ptr);

	void save_history(const char * filename);
	void load_history(const char * filename);

	// playback tracking
	void playback_on_stop();

	// playlist tracking
	void playlists_populate();
	void playlists_on_add(unsigned index);
	void playlists_on_delete(unsigned index);
	void playlists_on_reorder(const int * order, unsigned count);
	void playlists_on_switch(unsigned to);

	// metadb tracking
	void metadb_file_removed(metadb_handle * handle);
	void metadb_file_moved(metadb_handle * src, metadb_handle *dst);
	void metadb_info_changed(metadb_handle * handle);
};

int history_i::get_count() {
	insync2(cs);
	return data.get_count();
}

int history_i::get_index_from_offset(int offset) {
	insync2(cs);
	if (offset == 0) return now_playing;
	int index = (offset < 0) ? bound_prev + offset : bound_next + offset;
	if (index < 0 || index >= get_count()) index = -1;
	return index;
}

int history_i::get_offset_from_index(int index) {
	insync2(cs);
	if (index < bound_prev) return index - bound_prev;
	if (index > bound_next) return index - bound_next;
	return 0;
}

void history_i::add_callback(history_callback * ptr, bool callonadd) {
	if (ptr == NULL) return;

	insync2(cs);

	callbacks.add_item(ptr);

	if (callonadd) {
		ptr->on_items_added(0, get_count());
	}
}

void history_i::remove_callback(history_callback * ptr) {
	if (ptr == NULL) return;

	insync2(cs);

	int idx = callbacks.find_item(ptr);
	if (idx != -1) callbacks.remove_by_idx(idx);
}

void history_i::remove_mask(const bit_array & mask) {
	insync2(cs);

#ifdef VERBOSE_INDEX_ADJUSTMENT
	console::info("before remove_mask");
	console::info(string_printf("now_playing: %d, bound_prev: %d, bound_next: %d",
		now_playing,bound_prev,bound_next));
#endif

	// broadcast on_items_removing event
	if (initialized)
	{
		for (unsigned n = 0; n < callbacks.get_count(); n++)
		{
			history_callback * cb = callbacks[n];
			cb->on_items_removing(mask);
		}
	}

	unsigned count = data.get_count();

	// remove items
	data.delete_mask(mask);

	// adjust playback cursor
	if (mask[cursor]) cursor = get_count();
	else if (cursor >= 0)
	{
		cursor -= mask.calc_count(true, 0, cursor);
		if (cursor < 0) cursor = get_count();
	}

	// pre-adjust bound_prev
	bound_prev = mask.find(false, max(0,bound_prev-1), -max(0,bound_prev-1))+1;

	// count number of items deleted before bound_prev
	int val = mask.calc_count(true, 0, bound_prev);
	bound_prev -= val;
	if (now_playing < 0 || mask[now_playing]) {
		now_playing = -1;
		bound_next = bound_prev - 1;
	} 
	else {
		now_playing = bound_prev;
		bound_next = bound_prev;
	}

#ifdef VERBOSE_INDEX_ADJUSTMENT
	console::info("after remove_mask");
	console::info(string_printf("now_playing: %d, bound_prev: %d, bound_next: %d",
		now_playing,bound_prev,bound_next));
#endif

	if (initialized) {
		// broadcast on_items_removed event
		for (unsigned n = 0; n < callbacks.get_count(); n++) {
			history_callback * cb = callbacks[n];
			cb->on_items_removed(mask);
		}
	}
}

void history_i::get_data(ptr_list_interface<history_handle> & out) {
	insync2(cs);
	unsigned n, m = data.get_count();
	for (n = 0; n<m; ++n) {
		out.add_item(data.get_item(n)->handle_duplicate());
	}
}

void history_i::format_title(int index, string_base & out, const char * spec, const char * p_extra_items) {
	//insync2(cs);
	cs.enter();
	history_handle * handle = get_item_locked(index);
	if (handle != NULL) {
		extra_items_helper extra_items(p_extra_items);
		int offset = get_offset_from_index(index);
		if (offset < 0) {
			extra_items.add_item("history_prev", -offset);
		} 
		else if (offset > 0) {
			extra_items.add_item("history_next", offset);
		}
		else if (index == now_playing) {
			extra_items.add_item("history_is_playing", "1");
		}
		extra_items.add_item("history_number", index+1);

#ifdef _DEBUG
		if (index == cursor) {
			extra_items.add_item("history_cursor", "1");
		}
#endif
#if 0
		// SystemTimeToTzSpecificLocalTime is only available on WinNT or better
		GetSystemTime(&st_utc);
		SYSTEMTIME st_local;
		if (SystemTimeToTzSpecificLocalTime(0, &st_utc, &st_local)) {
			extra_items.add_item("history_timestamp", (const char *)string_print_date(st_local));
		}
		else {
			console::error("SystemTimeToTzSpezificLocalTime");
			console::info(string_print_last_error());
		}
#endif
		cs.leave();

		handle->handle_format_title(out, spec, extra_items);
	}
	else {
		cs.leave();
		out.reset();
	}
}

int history_i::get_playback_cursor() {
	insync2(cs);
	return cursor;
}

void history_i::set_playback_cursor(int index) {
	insync2(cs);
	if (index < 0) {
		cursor = -1;
	}
	else if (index >= get_count()) {
		cursor = get_count();
	}
	else {
		cursor = index;
	}
}

history_handle_i * history_i::get_item_locked(int index) {
	if (index >= 0 && index < get_count()) {
		return (history_handle_i *)data.get_item(index);
	}
	return 0;
}

history_handle * history_i::get_item(int index) {
	insync2(cs);
	history_handle * handle = get_item_locked(index);
	if (handle) handle->handle_add_ref();
	return handle;
}

static int find_in_playlist(metadb_handle * ptr, unsigned playlist = (unsigned)-1) {
	int idx = -1;
	metadb_handle_list data;
	if (playlist == (unsigned)-1) {
		playlist_oper::get()->get_data(data);
	} 
	else {
		playlist_switcher * p_switcher = playlist_switcher::get();
		if (playlist < p_switcher->get_num_playlists()) {
			p_switcher->get_playlist_data(playlist, data);
		}
	}
	unsigned n, count = data.get_count();
	for (n = 0; n < count; n++) {
		if (data[n] == ptr) {
			idx = n;
			break;
		}
	}
	data.delete_all();
	return idx;
}

int history_i::get_now_playing() {
	insync2(cs);
	return now_playing;
}

void history_i::play_item(int index) {
	//insync2(cs);
	/////////////////////////////
	// Enter critical section
	/////////////////////////////
	cs.enter();

	history_handle * handle = get_item_locked(index);
	if (handle != NULL) {
		int prev_playing = get_now_playing();
		now_playing = index;
		if (now_playing < 0) {
			// track is not in history
			// weird, get_item_locked should return NULL in that case...
			bound_next = bound_prev - 1;
		}
		else {
			bound_prev = now_playing;
			bound_next = now_playing;
		}
		set_playback_cursor(index+1);

		/////////////////////////////
		// Leave critical section
		/////////////////////////////
		cs.leave();

		playlist_handle * the_playlist = ((history_handle_i *)handle)->get_playlist();
		unsigned item_pl_index = the_playlist ? the_playlist->get_index() : (unsigned)-1;
		unsigned old_pl_index = pl_index;
		metadb_handle * track = handle->handle_get_track();
		{
			int playlist_index = find_in_playlist(track, item_pl_index);
			if (playlist_index >= 0) {
				if (item_pl_index != (unsigned)-1 && item_pl_index != pl_index) {
					playlist_switcher * p_switcher = playlist_switcher::get();
					p_switcher->set_active_playlist(item_pl_index);
					//p_switcher->reset_playing_playlist();
				}
				playlist_oper::get()->play_item(playlist_index);
				if (cfg_toggle_change_selection && config_var_int::g_get_value("CORE/Follow cursor")) {
					playlist_oper::get()->set_sel_one(playlist_index);
				}
			}
			else {
				play_control::get()->play_item(track);
			}
		}

		/////////////////////////////
		// Enter critical section
		/////////////////////////////
		cs.enter();

		fire_playing_track_changed(prev_playing, now_playing);
	}
	else {
		CONSOLE_WARNING("cannot change to a track that is not in the history");
	}
	/////////////////////////////
	// Leave critical section
	/////////////////////////////
	cs.leave();
}

void history_i::add_track(metadb_handle * track, const packed_date * timestamp, playlist_handle * p_playlist) {
	insync2(cs);
	if (track == 0) return;

	if (cfg_toggle_log_mode) {
		history_handle * temp = get_item_locked(now_playing);
		// check if track is already at the end of the log
		if (temp && temp->handle_get_track() == track) {
			bound_prev = now_playing;
			bound_next = now_playing;
			fire_playing_track_changed(now_playing, now_playing);
			return;
		}
	}
	else {
		history_handle * temp = get_item_locked(now_playing);
		// check if track is already playing
		if (temp && temp->handle_get_track() == track) {
			// should fix newest track not being seen as now playing at program start
			bound_prev = now_playing;
			bound_next = now_playing;
			fire_playing_track_changed(now_playing, now_playing);
			return;
		}

		int prev_playing = get_now_playing();

		// check if track is the one we expect to play next
		temp = get_item_locked(cursor);
		if (temp != NULL) {
			if (temp->handle_get_track() == track) {
				now_playing = cursor;
				// we know that now_playing >= 0;
				bound_prev = now_playing;
				bound_next = now_playing;

				set_playback_cursor(now_playing+1);
				fire_playing_track_changed(prev_playing, now_playing);
				return;
			}
			remove_mask(bit_array_range(cursor, data.get_count()-cursor));
		}
	}

	// check if we need to truncate the history
	if ((int)data.get_count() + 1 > cfg_max_size) {
		remove_mask(bit_array_range(0, data.get_count()-cfg_max_size+1));
	}

	history_handle * handle = create_handle(track, timestamp, p_playlist);

	if (cursor < 0) {
		now_playing = 0;
		data.insert_item(handle, 0);
	}
	else {
		now_playing = data.add_item(handle);
	}
	bound_prev = now_playing;
	bound_next = now_playing;
	set_playback_cursor(now_playing+1);

	fire_item_added(now_playing);
}

void history_i::add_track_autoplaylist(metadb_handle * track, const packed_date * timestamp) {
	insync2(cs);
	playlist_handle * p_playlist = 0;
	if (pl_index < playlists.get_count()) {
		p_playlist = playlists[pl_index];
	}
	add_track(track, timestamp, p_playlist);
}

history_handle_i * history_i::create_handle(metadb_handle * track, const packed_date * timestamp, playlist_handle * p_playlist) {
	history_handle_i * rv;

	if (timestamp == NULL) {
		SYSTEMTIME st;
		GetSystemTime(&st);
		packed_date timestamp(st);
		rv = new history_handle_i(track, timestamp, p_playlist);
	}
	else {
		rv = new history_handle_i(track, *timestamp, p_playlist);
	}

	return rv;
}

void history_i::fire_item_added(int index) {
	fire_items_added(index, 1);
}

void history_i::fire_items_added(int start, int count) {
	if (!initialized) return;
	CONSOLE_INFO(string_printf("added %i items to history at position %i (%i total)", count, start+1, get_count()));

	for (unsigned n = 0; n < callbacks.get_count(); n++) {
		callbacks[n]->on_items_added(start, count);
	}
}

void history_i::fire_playing_track_changed(int from, int to) {
	if (!initialized) return;
	CONSOLE_INFO(string_printf("changed history position from %i to %i (%i total)", from+1,to+1,get_count()));

	for (unsigned n = 0; n < callbacks.get_count(); n++) {
		callbacks[n]->on_playing_track_changed(from, to);
	}
}

void history_i::fire_item_modified(int index) {
	if (!initialized) return;
	CONSOLE_INFO(string_printf("modified item at position %i (%i total)", index+1, get_count()));

	for (unsigned n = 0; n < callbacks.get_count(); n++) {
		callbacks[n]->on_item_modified(index);
	}
}

void history_i::fire_item_replaced(int index) {
	if (!initialized) return;
	CONSOLE_INFO(string_printf("replaced item at position %i (%i total)", index+1, get_count()));

	for (unsigned n = 0; n < callbacks.get_count(); n++) {
		callbacks[n]->on_item_replaced(index);
	}
}

//////////////////////////////////////////////////
// history serialization
//////////////////////////////////////////////////

bool print_packed_date(const packed_date & val, string_base & out) {
	char buffer[32];

	int rv = sprintf(buffer, "%04d-%02d-%02d %02d:%02d:%02dZ",
		val.get_year(), val.get_month(), val.get_day(),
		val.get_hour(), val.get_minute(), val.get_second());
	out = buffer;

	return rv == 20;
}

bool scan_packed_date(packed_date & out, const char * ptr) {
	int year, month, day;
	int hour, minute, second;

	int rv = sscanf(ptr, "%04d-%02d-%02d %02d:%02d:%02dZ",
		&year, &month, &day, &hour, &minute, &second);
	if (rv != 6) return false;

	out.set_year(year);
	out.set_month(month);
	out.set_day(day);
	out.set_hour(hour);
	out.set_minute(minute);
	out.set_second(second);

	return true;
}


void history_i::save_history(const char * filename) {
	reader * r = file::g_open(filename, reader::MODE_WRITE_NEW);
	if (r != NULL) {
		r->write_guid_le(history_loader::guid_file_format_v1);

		unsigned count = get_count();
		r->write_dword_le(count);

		unsigned n;
		for (n=0; n<count; n++) {
			metadb_handle * handle = data.get_item(n)->handle_get_track();
			unsigned len = strlen(handle->handle_get_path());
			r->write_dword_le(len);
			if (len) r->write(handle->handle_get_path(), len);
			unsigned subsong = handle->handle_get_subsong_index();
			r->write_dword_le(subsong);

			string8 temp;
			packed_date timestamp;
			data.get_item(n)->handle_get_timestamp(timestamp);
			if (print_packed_date(timestamp, temp))
				len = temp.length();
			else
				len = 0;
			r->write_dword_le(len);
			if (len) r->write(temp, len);
		}

		r->reader_release();
	}
}

void history_i::load_history(const char * filename) {
	remove_all();

	reader * r = file::g_open(filename, reader::MODE_READ);
	if (r != NULL) {
		// try to get a buffered reader
		reader * r2 = reader_membuffer::create(r);
		if (r2 != NULL) {
			r->reader_release();
			r = r2;
		}

		GUID guid_read;
		r->read_guid_le(guid_read);

		if (guid_read == history_loader::guid_file_format_v1) {
			DWORD count = 0;
			r->read_dword_le(count);

			unsigned n;
			for (n=0; n<count; n++) {
				// read path
				string8 path;
				DWORD len;
				r->read_dword_le(len);
				if (len < 0 || len > 0x1000) {
					console::error(string_printf("invalid path length (0x%08x) while loading item %d of %d at offset 0x%08x",
						len, n+1, count, (unsigned int)r->get_position()));
					break;
				} else if (len > 0) {
					r->read(string_buffer(path,len),len);
				}

				// read subsong index
				DWORD subsong;
				r->read_dword_le(subsong);

				// read timestamp
				r->read_dword_le(len);
				if (len < 0 || len > 0x1000) {
					console::error(string_printf("invalid data length (0x%08x) while loading item %d of %d at offset 0x%08x",
						len, n+1, count, (unsigned int)r->get_position()));
					break;
				} else if (len > 0) {
					string8 temp;
					r->read(string_buffer(temp,len),len);
					packed_date timestamp;
					if (scan_packed_date(timestamp, temp)) {
						metadb_handle * handle = metadb::get()->handle_create(make_playable_location(path,subsong));
						if (handle != NULL) {
							add_track(handle, &timestamp);
							handle->handle_release();
						}

					}
				}
			}
		}

		r->reader_release();
	}

	initialized = true;
	bound_prev = bound_next + 1;
	fire_items_added(0, get_count());
}

//////////////////////////////////////////////////
// playback tracking
//////////////////////////////////////////////////

void history_i::playback_on_stop() {
	bound_prev = bound_next + 1;
	int prev_playing = now_playing;
	cursor = now_playing;
	now_playing = -1;
	fire_playing_track_changed(prev_playing, now_playing);
}

//////////////////////////////////////////////////
// playlist tracking
//////////////////////////////////////////////////

void history_i::playlists_populate() {
}

void history_i::playlists_on_add(unsigned index) {
	insync2(cs);
	playlists.insert_item(playlist_handle::create(), index);
	unsigned n, count = playlists.get_count();
	for (n = index; n < count; n++) {
		playlists[n]->set_index(n);
	}
	pl_index = playlist_switcher::get()->get_active_playlist();
}

void history_i::playlists_on_delete(unsigned index) {
	insync2(cs);
	if (index < playlists.get_count()) {
		playlist_handle * ptr = playlists.remove_by_idx(index);
		if (ptr) ptr->service_release();
	}
	pl_index = playlist_switcher::get()->get_active_playlist();
}

void history_i::playlists_on_reorder(const int * order, unsigned count) {
	insync2(cs);
	playlists.apply_order(order, count);
	pl_index = playlist_switcher::get()->get_active_playlist();
}

void history_i::playlists_on_switch(unsigned to) {
	insync2(cs);
	pl_index = to;
}

static service_factory_single_transparent_t< history, history_i > foo_history;

class play_callback_history : public play_callback {
public:
	virtual unsigned get_callback_mask() {return MASK_on_playback_new_track|MASK_on_playback_stop;}

	virtual void on_playback_starting() {}

	virtual void on_playback_new_track(metadb_handle *track) {
		foo_history.add_track_autoplaylist(track);
	}

	virtual void on_playback_stop(play_control::stop_reason reason) {
		if (reason == play_control::STOP_REASON_USER) {
			foo_history.playback_on_stop();
		}
	}

	virtual void on_playback_seek(double time) {}

	virtual void on_playback_pause(int state) {}

	virtual void on_playback_edited(metadb_handle * track) {};

	virtual void on_playback_dynamic_info(const file_info * info,bool b_track_change) {};
};

static play_callback_factory<play_callback_history> foo_play_callback;

class playlist_switcher_callback_history : public playlist_switcher_callback {
public:
	virtual void on_playlist_switch_before(unsigned from,unsigned to) {}

	virtual void on_playlist_switch_after(unsigned from,unsigned to) {
		foo_history.playlists_on_switch(to);
	}

	virtual void on_reorder(const int * order,unsigned count) {
		foo_history.playlists_on_reorder(order, count);
	}

	virtual void on_new_playlist(const char * name,unsigned idx,const ptr_list_interface<metadb_handle> & data) {
		foo_history.playlists_on_add(idx);
	}

	virtual void on_delete_playlist(unsigned idx) {
		foo_history.playlists_on_delete(idx);
	}

	virtual void on_rename_playlist(unsigned idx,const char * new_name) {}

	virtual void on_item_replaced(unsigned pls,unsigned item,metadb_handle * from,metadb_handle * to) {}
};

static playlist_switcher_callback_factory< playlist_switcher_callback_history > foo_playlist_switcher_callback;

//////////////////////////////////////////////////
// metadb change tracking
//////////////////////////////////////////////////

void history_i::metadb_file_removed(metadb_handle * handle) {
	insync2(cs);

	int n = 0;
	while (n < get_count()) {
		history_handle_i * ptr = get_item_locked(n);

		if (ptr->handle_get_track() == handle) {
			remove_item(n);
		}
		else {
			n++;
		}
	}
}

void history_i::metadb_file_moved(metadb_handle * src, metadb_handle * dst) {
	insync2(cs);

	const int count = get_count();
	for (int n = 0; n < count; n++) {
		history_handle_i * ptr = get_item_locked(n);

		if (ptr->handle_get_track() == src) {
			packed_date timestamp;
			ptr->handle_get_timestamp(timestamp);
			playlist_handle * playlist = ptr->get_playlist();
			history_handle_i * ptr2 = create_handle(dst, &timestamp, playlist);
			data.replace_item(n, ptr2);
			ptr->handle_release();
			fire_item_replaced(n);
		}
	}
}

void history_i::metadb_info_changed(metadb_handle * handle) {
	insync2(cs);

	const int count = get_count();
	for (int n = 0; n < count; n++) {
		history_handle_i * ptr = get_item_locked(n);

		if (ptr->handle_get_track() == handle) {
			fire_item_modified(n);
		}
	}
}

class metadb_callback_history : public metadb_callback {
public:
	virtual void on_info_edited(metadb_handle * handle) {
		foo_history.metadb_info_changed(handle);
	}

	virtual void on_info_reloaded(metadb_handle * handle) {
		foo_history.metadb_info_changed(handle);
	}

	virtual void on_file_removed(metadb_handle * handle) {
		foo_history.metadb_file_removed(handle);
	}

	virtual void on_file_moved(metadb_handle * src,metadb_handle * dst) {
		foo_history.metadb_file_moved(src, dst);
	}

	virtual void on_file_new(metadb_handle * handle) {}//reserved for future use

	virtual void on_file_removed_user(metadb_handle * handle) {
		foo_history.metadb_file_removed(handle);
	}
};

static service_factory_single_t< metadb_callback, metadb_callback_history > foo_metadb_callback;

//////////////////////////////////////////////////
// history initquit handler
//////////////////////////////////////////////////

class initquit_history : public initquit {
private:
	void get_history_path(string_base & out) {
		out.set_string(core_api::get_profile_path());
		out.add_string("history.flg");
	}

public:
	virtual void on_init() {
		if (cfg_toggle_save_history) {
			string8 path;
			get_history_path(path);
			foo_history.load_history(path);
		}
		initialized = true;
	}

	virtual void on_quit() {
		if (cfg_toggle_save_history) {
			string8 path;
			get_history_path(path);
			foo_history.save_history(path);
		}

		foo_history.remove_all();
	}
};

static initquit_factory< initquit_history > foo_initquit;
