#include "history_window.h"

#include "config.h"
#include "resource.h"

extern void show_help();

enum {
	/** Generic change, size is unchanged, but all items need to be redrawn. */
	MSG_HISTORY_CHANGED = WM_USER,
	/** History size changed. All items need to be redrawn. */
	MSG_HISTORY_SIZE_CHANGED,
	/** History item changed. Item at index wp needs to be redrawn. */
	MSG_HISTORY_ITEM_CHANGED,
	/** Reset active item to currently playing. */
	MSG_HISTORY_RESET_SELECTION,
	/** Destroy tooltip. */
	MSG_DESTROY_TOOLTIP,
	/** Custom font changed. Passes the new font in wParam. */
	MSG_FONT_CHANGED,
	/** Use system font setting changed. */
	MSG_FONT_SETTING_CHANGED,
	/** Custom color changed. */
	MSG_COLOR_CHANGED,
	/** Use system colors setting changed. */
	MSG_COLOR_SETTING_CHANGED,
	/** Display order setting changed. */
	MSG_DISPLAY_ORDER_CHANGED,
};

class cfg_var_callback_toggle : public cfg_var_notify {
	mem_block_list_t<HWND> g_wnd_list;
	UINT g_msg;
	WPARAM g_wp;

	virtual void on_var_change(const char * name, cfg_var * ptr) {
		int val = *(cfg_int *)ptr;

		unsigned n, count = g_wnd_list.get_count();
		for (n = 0; n < count; n++) {
			uSendMessage(g_wnd_list[n], g_msg, g_wp, val);
		}
	}

public:
	cfg_var_callback_toggle(cfg_var * var, UINT msg, WPARAM wp = 0) : cfg_var_notify(var), g_msg(msg), g_wp(wp) {}

	void add_window(HWND wnd) {
		g_wnd_list.add_item(wnd);
	}

	void remove_window(HWND wnd) {
		g_wnd_list.remove_item(wnd);
	}
};

static cfg_var_callback_toggle g_toggle_notify_sys_colors(&cfg_toggle_sys_colors, MSG_COLOR_SETTING_CHANGED);
static cfg_var_callback_toggle g_toggle_notify_sys_font(&cfg_toggle_sys_font, MSG_FONT_SETTING_CHANGED);

static cfg_var_callback_toggle g_toggle_notify_display_order(&cfg_toggle_ascending, MSG_DISPLAY_ORDER_CHANGED);


history_callback_ui::history_callback_ui(history_dialog_base * p_tgt) {
	tgt = p_tgt;
}

void history_callback_ui::on_items_added(int start, int count) {
	if (tgt != NULL) tgt->on_history_items_added(start, count);
}

void history_callback_ui::on_items_removing(const bit_array & mask) {
}

void history_callback_ui::on_items_removed(const bit_array & mask) {
	if (tgt != NULL) tgt->on_history_items_removed(mask);
}

void history_callback_ui::on_item_modified(int index) {
	if (tgt != NULL) tgt->on_history_item_modified(index);
}

void history_callback_ui::on_item_replaced(int index) {
	on_item_modified(index);
}

void history_callback_ui::on_playing_track_changed(int from, int to) {
	if (tgt != NULL) tgt->on_history_playing_track_changed(from, to);
}



void history_dialog_base::on_history_items_added(int start, int num) {
	uPostMessage(wnd, MSG_HISTORY_SIZE_CHANGED, 0, 0);
}

void history_dialog_base::on_history_items_removed(const bit_array & mask) {
	uPostMessage(wnd, MSG_HISTORY_SIZE_CHANGED, 0, 0);
}

void history_dialog_base::on_history_item_modified(int index) {
	uPostMessage(wnd, MSG_HISTORY_ITEM_CHANGED, index, 0);
}

void history_dialog_base::on_history_playing_track_changed(int from, int to) {
	uPostMessage(wnd, MSG_HISTORY_CHANGED, 0, 0);
}



history_dialog_base::history_dialog_base() {
	wnd_list = NULL;
	wnd_tooltip = NULL;
	count = 0;

	font_notify_view = NULL;
	color_notify_text = NULL;
	color_notify_bg = NULL;
	color_notify_sel_text = NULL;
	color_notify_sel_bg = NULL;

	history_notify = NULL;

	p_menu_manager = NULL;

	rc_tooltip.left=0;
	rc_tooltip.top=0;
	rc_tooltip.right=0;
	rc_tooltip.bottom=0;
	last_tooltip_index = -1;

	font_view = NULL;

	brush_bg = NULL;
	brush_sel_bg = NULL;
}


int history_dialog_base::h2ui(int idx_history)
{
	// parameter checking
	assert(idx_history >= -1);
	assert(idx_history < count);

	// conversion
	int idx_ui = idx_history;
	if (cfg_toggle_ascending && idx_history >= 0)
		idx_ui = count - idx_ui - 1;

#ifdef VERBOSE_INDEX_TRANSLATION
	console::info(string_printf("history: %d -> ui: %d (count: %d)",idx_history,idx_ui,count));
#endif

	// result checking
	assert(idx_ui >= -1);
	assert(idx_ui < count);

	return idx_ui;
}

int history_dialog_base::ui2h(int idx_ui)
{
	// parameter checking
	assert(idx_ui >= -1);
	assert(idx_ui < count);

	// conversion
	int idx_history = idx_ui;
	if (cfg_toggle_ascending && idx_history >= 0)
		idx_history = count - idx_history - 1;

#ifdef VERBOSE_INDEX_TRANSLATION
	console::info(string_printf("ui: %d -> history: %d (count: %d)",idx_ui,idx_history,count));
#endif

	// result checking
	assert(idx_history >= -1);
	assert(idx_history < count);

	return idx_history;
}

int history_dialog_base::get_focus()
{
	assert(wnd_list);

	int index = uSendMessage(wnd_list, LB_GETCURSEL, 0, 0);

	return ui2h(index);
}

void history_dialog_base::set_focus(int index) {
	assert(wnd_list);

	index = h2ui(index);

	uSendMessage(wnd_list, LB_SETCURSEL, index, 0);

	update_status_bar();
}

void history_dialog_base::set_count(int new_count) {
	//uSendMessage(wnd_list, WM_SETREDRAW, 0, 0);

	int index = -1;
	//index = get_focus();

	count = new_count;

	cache_resize(count);

	uSendMessage(wnd_list, LB_SETCOUNT, count, 0);

	if (index < 0) {
		index = history::get()->get_now_playing();
	}
	if (index >= 0) {
		set_focus(index);
	}

	//uSendMessage(wnd_list, WM_SETREDRAW, 1, 0);
}

void history_dialog_base::update_status_bar() {
	assert(wnd_list);

	history * h = history::get();
	if (h)
	{
		int index = get_focus();
		if (index >= 0)
		{
			string8 text;
			string8 spec;
			cfg_titleformat_statusbar.get_value(spec);
			h->format_title(index, text, spec, 0);
			set_status_text(text, false);
		}
	}
}

void history_dialog_base::update_list() {
	assert(wnd_list);

	uSendMessage(wnd_list, WM_SETREDRAW, 0, 0);

	cfg_titleformat_history.get_value(spec);

	RedrawWindow(wnd_list, NULL, NULL, RDW_INVALIDATE);

	uSendMessage(wnd_list, WM_SETREDRAW, 1, 0);

	set_caption_text(string_printf("History - %i entries (%s)", count, cfg_toggle_ascending ? "newest first" : "oldest first"));
}

void history_dialog_base::update_item(int index) {
	assert(wnd_list);

	cache_reset_item(index);

	uSendMessage(wnd_list, LVM_REDRAWITEMS, h2ui(index), h2ui(index)+1);
}

void history_dialog_base::play_item()
{
	int n = get_focus();
	history * h = history::get();
	if (n >= 0 && h != NULL)
	{
		h->play_item(n);
	}
}

void history_dialog_base::remove_item()
{
	int n = get_focus();
	history * h = history::get();
	if (n >= 0 && h != NULL)
	{
		h->remove_item(n);
	}
}

bool history_dialog_base::context_menu(HWND child)
{
	if (child != wnd_list) return false;
	int index = get_focus();
	if (index < 0) return true;

	history * h = history::get();
	history_handle * handle = h->get_item(index);
	if (!handle) return true;

	metadb_handle_list data;
	data.add_item(handle->handle_get_track());

	POINT p;
	GetCursorPos(&p);

	HMENU menu = CreatePopupMenu();

	p_menu_manager = menu_manager::create();

	uAppendMenu(menu, MF_STRING, ID_PLAY, "Play\tEnter");
	uAppendMenu(menu, MF_STRING, ID_REMOVE, "Remove from history\tDelete");
	uAppendMenu(menu, MF_STRING, ID_PREFS, "Settings...");
	uAppendMenu(menu, MF_STRING, ID_HELP, "Help");
	if (p_menu_manager)
	{
		p_menu_manager->init_context(data, 0);

		if (p_menu_manager->get_root() != 0)
			uAppendMenu(menu, MF_SEPARATOR, 0, 0);

		p_menu_manager->v_win32_build_menu(menu, ID_CUSTOM_BASE, -1);
		p_menu_manager->v_win32_auto_mnemonics(menu);
	}

	int cmd = TrackPopupMenu(menu, TPM_RIGHTBUTTON | TPM_NONOTIFY | TPM_RETURNCMD,
		p.x, p.y, 0, wnd, 0);

	if (cmd == ID_PLAY)
	{
		play_item();
	}
	else if (cmd == ID_REMOVE)
	{
		remove_item();
	}
	else if (cmd == ID_PREFS)
	{
		play_control::get()->show_config("History");
	}
	else if (cmd == ID_HELP)
	{
		show_help();
	}
	else if (cmd >= ID_CUSTOM_BASE)
	{
		p_menu_manager->execute_by_id(cmd - ID_CUSTOM_BASE);
	}

	if (p_menu_manager) p_menu_manager->service_release();
	p_menu_manager = NULL;

	handle->handle_release();

	DestroyMenu(menu);

	return true;
}

void history_dialog_base::destroy_tooltip()//destroy the tooltip
{
	if (wnd_tooltip) {
		DestroyWindow(wnd_tooltip);
		wnd_tooltip=0;
	}
	last_tooltip_index = -1;
}

BOOL history_dialog_base::create_tooltip(const char * text)
{
	
	destroy_tooltip();
	
	bool update = false;
	
	wnd_tooltip = uCreateWindowEx(WS_EX_TOPMOST, TOOLTIPS_CLASS, NULL,
		WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP | TTS_NOPREFIX,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		wnd_list, 0, core_api::get_my_instance(), NULL);
	
    SetWindowPos(wnd_tooltip, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
	
	RECT rect;
	GetClientRect (wnd_list,  &rect);
	
	uTOOLINFO ti;
	memset(&ti,0,sizeof(ti));
	
	ti.cbSize = sizeof(uTOOLINFO);
	ti.uFlags = TTF_SUBCLASS|TTF_TRANSPARENT;
	ti.hwnd = wnd_list;  //we could use wnd_host, but then we have to relay mouse messages to the tooltip control instead of letting it subclass the listbox, or something like that
	ti.hinst = core_api::get_my_instance();
	ti.uId = IDC_TOOLTIP;
	ti.lpszText = const_cast<char *>(text);
	ti.rect = rect;
	
	return uToolTip_AddTool(wnd_tooltip, &ti, update);
}

HFONT history_dialog_base::get_font() {
	if (cfg_toggle_sys_font) {
		return (HFONT)GetStockObject(DEFAULT_GUI_FONT);
	} else {
		return font_view;
	}
}

COLORREF history_dialog_base::get_color(int index) {
	COLORREF rv = 0x000000;

	switch (index) {

	case COLOR_WINDOW:
		{
			rv = cfg_toggle_sys_colors
				? GetSysColor(COLOR_WINDOW)
				: cfg_color_bg;
		}
		break;

	case COLOR_HIGHLIGHT:
		{
			rv = cfg_toggle_sys_colors
				? GetSysColor(COLOR_HIGHLIGHT)
				: cfg_color_sel_bg;
		}
		break;

	case COLOR_WINDOWTEXT:
		{
			rv = cfg_toggle_sys_colors
				? GetSysColor(COLOR_WINDOWTEXT)
				: cfg_color_text;
		}
		break;

	case COLOR_HIGHLIGHTTEXT:
		{
			rv = cfg_toggle_sys_colors
				? GetSysColor(COLOR_HIGHLIGHTTEXT)
				: cfg_color_sel_text;
		}
		break;

	}

	return rv;
}

HBRUSH history_dialog_base::get_brush(int index) {
	HBRUSH rv = NULL;

	switch (index) {

	case COLOR_WINDOW:
		{
			if (!brush_bg) {
				brush_bg = CreateSolidBrush(get_color(index));
			}
			rv = brush_bg;
		}
		break;

	case COLOR_HIGHLIGHT:
		{
			if (!brush_sel_bg) {
				brush_sel_bg = CreateSolidBrush(get_color(index));
			}
			rv = brush_sel_bg;
		}
		break;
	}

	return rv;
}

void history_dialog_base::flush_brushes() {
	if (brush_bg) DeleteObject(brush_bg);
	brush_bg = NULL;
	if (brush_sel_bg) DeleteObject(brush_sel_bg);
	brush_sel_bg = NULL;
}

BOOL history_dialog_base::on_ctlcolorlistbox(WPARAM wp, LPARAM lp) {
	return (BOOL)get_brush(COLOR_WINDOW);
}

BOOL history_dialog_base::on_initdialog(WPARAM wp, LPARAM lp) {
	// get valid titleformat script, do this before the ListBox attempts drawing
	cfg_titleformat_history.get_value(spec);

	// initialize list window
	wnd_list = uCreateWindowEx(WS_EX_CLIENTEDGE,
		WC_LISTBOX, "",
		WS_CHILD | WS_VISIBLE | WS_VSCROLL | WS_TABSTOP |
		LBS_NOTIFY | LBS_NOINTEGRALHEIGHT | LBS_OWNERDRAWFIXED | LBS_NODATA,
		0, 0, 200, 500,
		wnd, (HMENU)IDC_LIST, core_api::get_my_instance(), NULL);
	//subclass the listbox
	ListProcPtr = (WNDPROC)uSetWindowLong(wnd_list, GWL_WNDPROC, (LPARAM)ListHookProc);
	
	uSendMessage(wnd_list, WM_SETFONT, (WPARAM)get_font(), MAKELPARAM(0, 0));

	// give focus to list window
	SetFocus(wnd_list);

	// register font notification callback
	font_notify_view = new config_var_callback_font(&cfg_font_history, wnd, MSG_FONT_CHANGED);

	// register color notification callbacks
	color_notify_text = new config_var_callback_color(&cfg_color_text, wnd, MSG_COLOR_CHANGED, COLOR_WINDOWTEXT);
	color_notify_bg = new config_var_callback_color(&cfg_color_bg, wnd, MSG_COLOR_CHANGED, COLOR_WINDOW);
	color_notify_sel_text = new config_var_callback_color(&cfg_color_sel_text, wnd, MSG_COLOR_CHANGED, COLOR_HIGHLIGHTTEXT);
	color_notify_sel_bg = new config_var_callback_color(&cfg_color_sel_bg, wnd, MSG_COLOR_CHANGED, COLOR_HIGHLIGHT);

	// register color setting callback
	g_toggle_notify_sys_colors.add_window(wnd);

	// register font setting callback
	g_toggle_notify_sys_font.add_window(wnd);

	// register display order callback
	g_toggle_notify_display_order.add_window(wnd);

	// register history callback
	history_notify = new history_callback_ui(this);
	history::get()->add_callback(history_notify, true);

	// reset selection to playing item
	// post message, so it is processed after posted history size message from history callback
	uPostMessage(wnd, MSG_HISTORY_RESET_SELECTION, 0, 0);

	// update dialog layout
	layout();

	return TRUE;
}

BOOL history_dialog_base::on_destroy(WPARAM wp, LPARAM lp) {
	uSetWindowLong(wnd_list, GWL_WNDPROC, (LPARAM)ListProcPtr);

	history::get()->remove_callback(history_notify);

	delete history_notify;

	delete font_notify_view;
	delete color_notify_text;
	delete color_notify_bg;
	delete color_notify_sel_text;
	delete color_notify_sel_bg;

	g_toggle_notify_sys_colors.remove_window(wnd);
	g_toggle_notify_sys_font.remove_window(wnd);
	g_toggle_notify_display_order.remove_window(wnd);

	cache_resize(0);

	flush_brushes();

	return TRUE;
}

BOOL history_dialog_base::on_command(WPARAM wp, LPARAM lp) {
	BOOL rv = FALSE;

	switch (wp) {
		
	case IDOK:
		{
			play_item();
			rv = TRUE;
		}
		break;

	case IDC_LIST | (LBN_SELCHANGE << 16):
		{
			update_status_bar();
		}
		break;

	case IDC_LIST | (LBN_DBLCLK << 16):
		{
			play_item();
		}
		break;
	}
	return rv;
}

BOOL history_dialog_base::on_notify(WPARAM wp, LPARAM lp) {
	return FALSE;
}

void history_dialog_base::layout() {
	RECT rc;
	GetClientRect(wnd, &rc);
	SetWindowPos(wnd_list, NULL, 0, 0, rc.right, rc.bottom, SWP_NOZORDER|SWP_NOMOVE);
}

BOOL history_dialog_base::on_size(WPARAM wp, LPARAM lp) {
	layout();

	return TRUE;
}

void history_dialog_base::draw_item(int index, HDC dc, RECT rc, bool selected, bool focus) {
	HBRUSH br;
	// get background brush
	br = get_brush(selected ? COLOR_HIGHLIGHT : COLOR_WINDOW);

	// fill item background
	FillRect(dc, &rc, br);

	// draw focus rectangle
	if (cfg_toggle_sys_frame) {
		// see ODA_FOCUS handler below		
	} else {
		// focus is control focus, selected is item selection
		if (focus) {
			HBRUSH br_focus = CreateSolidBrush(cfg_color_frame);
			FrameRect(dc, &rc, br_focus);//draw frame
			DeleteObject(br_focus);
		}
	}

	HFONT old_font = (HFONT)SelectObject(dc, get_font());

	const char * text = "";

	//retrieve string from cache
	if (index >= 0 && index < count) {
		text = cache_get(ui2h(index));
	}

	//render text
	if (text != NULL) {
		uTextOutColorsTabbed(dc,
			text,
			strlen(text),
			&rc, 2, &rc,
			selected, selected ? (get_color(COLOR_HIGHLIGHTTEXT) ^ 0xFFFFFF) : get_color(COLOR_WINDOWTEXT), 0);
	} else {
		// weird, cache size and view size are out of sync
#ifdef _DEBUG
		console::info(string_printf("view size: %d, cache size: %", count, cache.get_size()));
#endif
	}

	SelectObject(dc, old_font);
}

BOOL history_dialog_base::on_drawitem(WPARAM wp, LPARAM lp) {
	LPDRAWITEMSTRUCT lpdis = (LPDRAWITEMSTRUCT) lp;
	if (lpdis->CtlID == IDC_LIST)
	{
		if (lpdis->itemID >= 0) 
		{ 
			//device context to draw upon to
			HDC dc = lpdis->hDC;
			//item rect
			RECT rc = lpdis->rcItem;
			
			switch (lpdis->itemAction) 
			{ 
			case ODA_SELECT: 
			case ODA_DRAWENTIRE: 
				{
					bool selected = ((lpdis->itemState & ODS_SELECTED) !=0);
					bool focus = (GetFocus() == lpdis->hwndItem) && ((lpdis->itemState & ODS_FOCUS) !=0);

					// thanks to fixed resizing code, double buffering should not be necessary
					if (false /* use_double_buffering */) {
						HDC mem_dc = CreateCompatibleDC(dc);
						HBITMAP mem_bmp = CreateCompatibleBitmap(dc, rc.right-rc.left, rc.bottom-rc.top);
						HBITMAP old_bmp = (HBITMAP)SelectObject(mem_dc, mem_bmp);

						HFONT old_font = (HFONT)SelectObject(mem_dc, get_font());

						RECT mem_rc;
						mem_rc.left = 0;
						mem_rc.top = 0;
						mem_rc.right = rc.right-rc.left;
						mem_rc.bottom = rc.bottom-rc.top;

						draw_item(lpdis->itemID, mem_dc, mem_rc, selected, focus);
						
						BitBlt(dc, rc.left, rc.top, rc.right-rc.left, rc.bottom-rc.top,
							mem_dc, 0, 0, SRCCOPY);

						SelectObject(mem_dc, old_font);

						SelectObject(mem_dc, old_bmp);
						DeleteObject(mem_bmp);
						DeleteDC(mem_dc);
					} else {
						draw_item(lpdis->itemID, dc, rc, selected, focus);
					}
				}
				break; 
				
				//meh
			case ODA_FOCUS:
				{
					bool focus = ((lpdis->itemState & ODS_FOCUS) != 0);
					//console::info(string_printf("item %d %s focus", lpdis->itemID, focus ? "gains" : "loses"));

					// draw focus rectangle
					if (cfg_toggle_sys_frame) {
						// same function for drawing and removing focus rectangle
						DrawFocusRect(dc, &rc);
					} else {
						// drawn in draw_item()
						int idx = uSendMessage(wnd_list, LB_GETCURSEL, 0, 0);
						if (lpdis->itemID >= 0) {
							draw_item(lpdis->itemID, lpdis->hDC, lpdis->rcItem,  lpdis->itemState & ODS_SELECTED, focus);
						}
					}
				}
				break; 
			} 
		}
	}
	return TRUE;
}

int history_dialog_base::get_item_height() {
	int rv = 1;
	if (wnd_list) {
		HFONT font = (HFONT)uSendMessage(wnd_list, WM_GETFONT, 0, 0 );

		rv = uGetFontHeight(font);
		if (rv < 1) rv = 1; //ahem avoid divide-by-zero mess
	}
	return rv;
}

BOOL history_dialog_base::on_measureitem(WPARAM wp, LPARAM lp) {
	LPMEASUREITEMSTRUCT lpmis = (LPMEASUREITEMSTRUCT) lp;

	if(lpmis->CtlID == IDC_LIST)
	{
		lpmis->itemHeight = get_item_height(); 
		return TRUE;
	}

	return FALSE;
}

BOOL history_dialog_base::on_settingchange(WPARAM wp, LPARAM lp) {
	flush_brushes();

	return TRUE;
}

BOOL history_dialog_base::on_keydown(WPARAM wp, LPARAM lp) {
	BOOL rv = FALSE;

#define IsKeyPressed(vk) ((GetKeyState(vk) & 0x8000) != 0)
	bool shift_pressed = IsKeyPressed(VK_SHIFT);
	bool ctrl_pressed = IsKeyPressed(VK_CONTROL);
	bool alt_pressed = IsKeyPressed(VK_MENU);
#undef IsKeyPressed

	if (wp == VK_DELETE && !(shift_pressed || ctrl_pressed || alt_pressed)) {
		history::get()->remove_item(get_focus());
		rv = TRUE;
	} else {
		rv = keyboard_shortcut_manager::get()->on_keydown_auto(wp);
	}
	if (!rv && wp == VK_TAB) {
		rv = TRUE;
		ui_extension::g_on_tab(wnd_list);
	}
	return rv;
}

BOOL history_dialog_base::on_syskeydown(WPARAM wp, LPARAM lp) {
	BOOL rv = !keyboard_shortcut_manager::get()->on_keydown_auto(wp);
	return rv;
}

BOOL history_dialog_base::on_message(UINT msg, WPARAM wp, LPARAM lp)
{
	switch(msg)
	{
	case WM_INITDIALOG:
		return on_initdialog(wp, lp);

	case WM_DESTROY:
		return on_destroy(wp, lp);

	case WM_SIZE:
		return on_size(wp, lp);

	case WM_COMMAND:
		return on_command(wp, lp);

	case WM_CONTEXTMENU:
		return context_menu((HWND)wp);

	case WM_MENUSELECT:
		return on_menuselect(wp, lp);

	case WM_NOTIFY:
		return on_notify(wp, lp);

	case WM_MEASUREITEM:
		return on_measureitem(wp, lp);

	case WM_DRAWITEM:
		return on_drawitem(wp, lp);

	case WM_CTLCOLORLISTBOX:
		return on_ctlcolorlistbox(wp, lp);

	case WM_SETTINGCHANGE:
		return on_settingchange(wp, lp);

	case WM_VKEYTOITEM:
		{
			if (wnd_list == (HWND)lp) {
				if (keyboard_shortcut_manager::get()->on_keydown_auto(LOWORD(wp))) {
					return -2;
				} else {
					return -1;
				}
			}
		}
		break;

	case MSG_HISTORY_CHANGED:
		{
			cache_reset();
			update_list();
		}
		break;

	case MSG_HISTORY_SIZE_CHANGED:
		{
			set_count(history::get()->get_count());

			update_list();
		}
		break;

	case MSG_HISTORY_ITEM_CHANGED:
		{
			update_item(wp);
		}
		break;

	case MSG_HISTORY_RESET_SELECTION:
		{
			int index = history::get()->get_now_playing();
			if (index >= 0)
				set_focus(index);
		}
		break;

	case MSG_FONT_CHANGED:
		{
			font_view = (HFONT)wp;
			if (!cfg_toggle_sys_font) {
				uSendMessage(wnd_list, WM_SETFONT, wp, lp);
			}
			uSendMessage(wnd_list, LB_SETITEMHEIGHT, 0, MAKELPARAM(get_item_height(), 0));
		}
		break;

	case MSG_FONT_SETTING_CHANGED:
		{
			uSendMessage(wnd_list, WM_SETFONT, (WPARAM)get_font(), MAKELPARAM(1, 0));
			uSendMessage(wnd_list, LB_SETITEMHEIGHT, 0, MAKELPARAM(get_item_height(), 0));
		}
		break;

	case MSG_COLOR_CHANGED:
		{
			if (!cfg_toggle_sys_colors) {
				flush_brushes();
				RedrawWindow(wnd_list, NULL, NULL, RDW_INVALIDATE);
			}
		}
		break;

	case MSG_COLOR_SETTING_CHANGED:
		{
			flush_brushes();
			RedrawWindow(wnd_list, NULL, NULL, RDW_INVALIDATE);
		}
		break;

	case MSG_DISPLAY_ORDER_CHANGED:
		{
			update_list();
		}
		break;

	default:
		return FALSE;
	}
	return TRUE;
}

LRESULT history_dialog_base::list_hook(UINT msg, WPARAM wp, LPARAM lp) {
	switch (msg) {

	case WM_GETDLGCODE:
		{
			return DLGC_WANTALLKEYS;
		}
		break;

	case WM_KEYDOWN:
		{
			int rv = on_keydown(wp, lp);
			if (!rv) {
				rv = TRUE;
				if (wp == VK_RETURN) {
					uSendMessage(wnd, WM_COMMAND, IDOK, (LPARAM)wnd_list);
				} else if (wp == VK_ESCAPE) {
					uSendMessage(wnd, WM_COMMAND, IDCANCEL, (LPARAM)wnd_list);
				} else {
					rv = FALSE;
				}
			}
			if (rv) return 0;
		}
		break;

	case WM_SYSKEYDOWN:
		{
			int rv = on_syskeydown(wp, lp);
		}
		break;

	case WM_CHAR:
		{
			// disable processing of char messages
			return 0;
		}
		break;

	case WM_SIZE:
		{
			LRESULT rv = uCallWindowProc(ListProcPtr, wnd_list, msg, wp, lp);
			RedrawWindow(wnd_list, NULL, NULL, RDW_INVALIDATE);
			return rv;
		}
		break;

	case WM_RBUTTONDOWN:
		{
			int rv = uSendMessage(wnd_list, LB_ITEMFROMPOINT, 0, lp);
			if (HIWORD(rv) == 0) {
				int idx = LOWORD(rv);
				uSendMessage(wnd_list, LB_SETCURSEL, idx, 0);
				SetFocus(wnd_list);
			}
		}
		break;

	case WM_MOUSEMOVE:
		if (1 && !(wp & MK_LBUTTON))//cfg_playlist_sidebar_tooltips
		{
			int rv = uSendMessage(wnd_list, LB_ITEMFROMPOINT, 0, lp);
			int idx = LOWORD(rv);
			if (idx >= 0 && HIWORD(rv) == 0 && last_tooltip_index != idx)
			{
				RECT item;
				//get item rect
				uSendMessage(wnd_list, LB_GETITEMRECT, idx, (LPARAM)&item);

				//get handle to device context
				HDC dc = GetDC(wnd);

				const char * temp = "";
				if (idx < count) {
					temp = cache_get(ui2h(idx));
				}

				string8 text;
				titleformat::remove_color_marks(temp, text);

				//get handle for font
				HFONT fnt = (HFONT)uSendMessage(wnd_list, WM_GETFONT, 0, 0);

				//select it into the dc, as it has teh system font selected into it right now
				HFONT fnt_old = (HFONT)SelectObject(dc, fnt);
				SIZE size;
				uGetTextExtentPoint32(dc, text, text.length(), &size);
				int width = size.cx;
				SelectObject(dc, fnt_old);//select the old font back in
				ReleaseDC(wnd_list, dc);

				 // is clipped?
				if (width + 4 > item.right-item.left)
				{
					//store item rect
					rc_tooltip = item;
					//create tooltip
					create_tooltip(text);
				}
				else if (wnd_tooltip) {
					//otherwise destroy the old tooltip
					destroy_tooltip();
				}
				
			}
			else if (HIWORD(rv) && wnd_tooltip) {
				destroy_tooltip();
			}
			last_tooltip_index = (HIWORD(rv) == 0) ? idx : -1;
		}
		else
		{
			if (wnd_tooltip) {
				destroy_tooltip();
			}
		}
		break;
		
	case WM_NOTIFY:
		{
			switch (((LPNMHDR)lp)->idFrom)
			{
			case IDC_TOOLTIP:
				switch (((LPNMHDR)lp)->code)
				{
				case TTN_SHOW:
					{
					
					//position tooltip over item

					RECT rc, rc_tool;
					
					rc = rc_tooltip;
					ClientToScreen(wnd_list, (LPPOINT)&rc);

					GetWindowRect(wnd_tooltip, &rc_tool);

					int offset = MulDiv(get_item_height() - rc_tool.bottom+rc_tool.top, 1, 2);
					
					rc.top += offset;
					
#if 0
					// brings window into foreground, even if other window is (still) active
					SetWindowPos(wnd_tooltip,
						NULL,
						rc.left, rc.top,
						0, 0,
						SWP_NOSIZE|SWP_NOZORDER|SWP_NOACTIVATE);
#endif
					}
					
					return TRUE;
				}
				case TTN_POP:
					{
						POINT pt;
						GetCursorPos(&pt);
						RECT rc;
						GetWindowRect(wnd, &rc);
						if (!PtInRect(&rc, pt)) uPostMessage(wnd, MSG_DESTROY_TOOLTIP, 0, 0);
					}
					break;
				break;
			}
		}
		break;
	}

	return uCallWindowProc(ListProcPtr, wnd_list, msg, wp, lp);
}

LRESULT WINAPI history_dialog_base::ListHookProc(HWND wnd, UINT msg, WPARAM wp, LPARAM lp) {
	history_dialog_base * _this = reinterpret_cast<history_dialog_base *>(uGetWindowLong(GetParent(wnd), DWL_USER));

	return _this != NULL ? _this->list_hook(msg, wp, lp) : uDefWindowProc(wnd, msg, wp, lp);
}

BOOL history_dialog_base::on_menuselect(WPARAM wp, LPARAM lp) {
	unsigned item  = (UINT) LOWORD(wp);
	unsigned flags = (UINT) HIWORD(wp);
	HMENU menu     = (HMENU) lp;

	string8 text;
	if (flags == 0xffff) {
	} else if (flags & MF_POPUP) {
	} else if (item == ID_PREFS) {
		text = "Opens History preferences dialog.";
	} else if (item == ID_PLAY) {
		text = "Plays selected track.";
	} else if (item == ID_REMOVE) {
		text = "Removes selected track from history.";
	} else if (item >= ID_CUSTOM_BASE && p_menu_manager != NULL) {
		menu_node * node = p_menu_manager->find_by_id(item - ID_CUSTOM_BASE);
		if (node != NULL) node->get_description(text);
	}

	set_status_text(text, true);

	return TRUE;
}

const char * history_dialog_base::cache_get(history_index_t index) {
	if ((unsigned)index < cache.get_size()) {
		if (cache[index] == 0)
		{
			string8_fastalloc temp;
			history::get()->format_title(index, temp, spec, 0);
			cache[index] = strdup(temp);
		}
		return cache[index];
	}
	return 0;
}

void history_dialog_base::cache_reset() {
	unsigned n, count = cache.get_size();
	for (n = 0; n < count; n++) {
		free(cache[n]);
		cache[n] = 0;
	}
}

void history_dialog_base::cache_reset_item(history_index_t index) {
	if ((unsigned)index < cache.get_size()) {
		free(cache[index]);
		cache[index] = 0;
	}
}

void history_dialog_base::cache_resize(size_t count) {
	cache_reset();
	cache.set_size(count);
	cache.zeromemory();
}

